/*
  If Arduino communicates with both the W5100 and
  SD card using the SPI bus (through the ICSP header).
  This is on digital pins 10, 11, 12, and 13 on the Uno
  and pins 50, 51, and 52 on the Mega. On both boards,
  pin 10 is used to select the W5100 and pin 4 for the SD
  card. These pins cannot be used for general I/O. On the
  Mega, the hardware SS pin, 53, is not used to select either
  the W5100 or the SD card, but it must be kept as an output
  or the SPI interface won't work.
*/

#include <Servo.h>
#include <Ethernet.h>
#include <Wire.h>
#include <EthernetUdp.h>
#include <stdio.h>

IPAddress ip(192, 168, 10, 240);
byte mac[] = {0x2C, 0xF7, 0xF1, 0x08, 0x28, 0xF1};
EthernetUDP udp_server;
#define UDP_TX_PACKET_MAX_SIZE 860
char packet_buffer[UDP_TX_PACKET_MAX_SIZE];

//Global_______________________________________________
long prevT = 0;
float eprev1 = 0;
float eprev2 = 0;
float eprev3 = 0;
float eprev4 = 0;
float eprev5 = 0;
float eprev6 =  0;
float eprevTalus = 0;
float eprevMetatarsal = 0;
float eintegral1 = 0;
float eintegral2 = 0;
float eintegral3 = 0;
float eintegral4 = 0;
float eintegral5 = 0;
float eintegral6 = 0;
float eintegralTalus = 0;
float eintegralMetatarsal = 0;
float power_force = 0;
int servo = 0;
int servoPos[6];
unsigned long interval = 0;
unsigned long interval2 = 0;
bool oneshot = true;
float baselineOutput = 255; // Baseline sensor value
float sensorRating = 45; // Sensors rating in newton
byte i2cAddress1 = 0x07; // Slave address talus (SingleTact), default 0x04
byte i2cAddress2 = 0x08; // Slave address (SingleTact), default 0x04
bool startTalus = false;

// Force set-points
int force_talus = 15;
int force_metatarsal = 6;

// PID-values platform
float kp_platform = 25;
float ki_platform = 7;
float kd_platform = 0;

// PID-values Talus
float kp_talus_pos = 2.5;
float ki_talus_pos = 0.02;
float kd_talus_pos = 0.01;

float kp_talus_force = 25;
float ki_talus_force = 0.02;
float kd_talus_force = 0.01;


// PID-values metatarsal
float kp_metatarsal = 25;
float ki_metatarsal = 7;
float kd_metatarsal = 0;


// PINS________________________________________________
//Servo 1
#define PWM1 5       //motor PWM pin
#define IN21 30        //motor controller pin2
#define IN11 31       //motor controller pin1
#define POT1 A0      //motor potentiometer
//Servo 2
#define PWM2 6       //motor PWM pin
#define IN22 32       //motor controller pin2
#define IN12 33       //motor controller pin1
#define POT2 A1      //motor potentiometer
//Servo 3
#define PWM3 7  // motor PWM pin
#define IN23 34  // motor controller pin2
#define IN13 35  // motor controller pin1
#define POT3 A12 // motor potentiometer
//Servo 4
#define PWM4 8  // motor PWM pin
#define IN24 36  // motor controller pin2
#define IN14 37  // motor controller pin1
#define POT4 A13 // motor potentiometer
//Servo 5
#define PWM5 9  // motor PWM pin
#define IN25 38  // motor controller pin2
#define IN15 39  // motor controller pin1
#define POT5 A14 // motor potentiometer
//Servo 6
#define PWM6 12  // motor PWM pin
#define IN26 40  // motor controller pin2
#define IN16 41  // motor controller pin1
#define POT6 A15 // motor potentiometer
//Talus servo
#define PWM7 13  // motor PWM pin
#define IN27 1  // motor controller pin2
#define IN17 53  // motor controller pin1
#define POT7 A2 // motor potentiometer

int IN1[] = {IN11, IN12, IN13, IN14, IN15, IN16};
int IN2[] = {IN21, IN22, IN23, IN24, IN25, IN26};
int PWM[] = {PWM1, PWM2, PWM3, PWM4, PWM5, PWM6};
int pot[] = {POT1, POT2, POT3, POT4, POT5, POT6};
float eprevList[] = {eprev1, eprev2, eprev3, eprev4, eprev5, eprev6};
float eintegralList[] = {eintegral1, eintegral2, eintegral3, eintegral4,
eintegral5, eintegral6};
// Setup
void
setup()
{
  // Servo 1
  pinMode(PWM1, OUTPUT);
  pinMode(IN11, OUTPUT);
  pinMode(IN21, OUTPUT);
  //Servo 2
  pinMode(PWM2, OUTPUT);
  pinMode(IN22, OUTPUT);
  pinMode(IN12, OUTPUT);
  // Servo 3
  pinMode(PWM3, OUTPUT);
  pinMode(IN23, OUTPUT);
  pinMode(IN13, OUTPUT);
  // Servo 4
  pinMode(PWM4, OUTPUT);
  pinMode(IN24, OUTPUT);
  pinMode(IN14, OUTPUT);
  // Servo 5
  pinMode(PWM5, OUTPUT);
  pinMode(IN25, OUTPUT);
  pinMode(IN15, OUTPUT);
  // Servo 6
  pinMode(PWM6, OUTPUT);
  pinMode(IN26, OUTPUT);
  pinMode(IN16, OUTPUT);
  // Talus servo
  pinMode(PWM7, OUTPUT);
  pinMode(IN27, OUTPUT);
  pinMode(IN17, OUTPUT);
  Wire.begin();
  Serial.begin(57600);
  Ethernet.begin(mac, ip);
  delay(500);
  udp_server.begin(8888);
}

// Read from UDP-server______________________________________________
String readUDP()
{
  int packet_size = udp_server.parsePacket();
  if (packet_size)
  {
    for (int i = 0; i < UDP_TX_PACKET_MAX_SIZE; i++) packet_buffer[i] = 0;
    udp_server.read(packet_buffer, UDP_TX_PACKET_MAX_SIZE);
    Serial.println(packet_buffer);

    return String(packet_buffer);
  }
  else
  {
    return "0";
  }
}

// Send to UDP-server____________________________________________________________
void sendUDP(String message) {
  udp_server.beginPacket(IPAddress(192, 168, 10, 200), 60028);
  udp_server.write(message.c_str(), message.length());
  udp_server.endPacket();
}

// Read potentiometer value from the actuator____________________________________
int readPotentiometer(uint8_t POT)
{
  int sensorValue = analogRead(POT);
  return sensorValue;
}

// Calculate control signal__________________________________________________
float calculateU(int target, int pos, float kp, float ki, float kd, float& eprev,
float& eintegral)
{
  // PID-controller
  //time diference
  long currT = micros();
  float deltaT = ((float) (currT - prevT)) / (1.0e6);


  //Error calculation
  int e = target - pos;
  //PID Calculation
  float dedt = (e - eprev) / (deltaT); //Derivative
  eintegral = eintegral + e * deltaT; //Integral
  float u = kp * e + kd * dedt + ki * eintegral; //Control signal

  eprev = e;
  prevT = currT;
  return u;
}

//MOTOR____________________________________________________________________
void setMotor(float u, int pwm, int in1, int in2) {
  //Motor power
  float pwr = fabs(u);  //fabs == floating point, absolute value
  if (pwr > 255) {
    pwr = 255;
  }
  // Motor direction
  int dir = 1;
  if (u < 0)
  {

    dir = -1; //if the control signal is negative, we flip the motor direction
  }
  if (pwr > 40)
  {
    analogWrite(pwm, pwr);
    if (dir == 1) {
      digitalWrite(in1, HIGH);
      digitalWrite(in2, LOW);
    }
    else if (dir == -1) {
      digitalWrite(in1, LOW);
      digitalWrite(in2, HIGH);

    }
    else {
      digitalWrite(in1, LOW);
      digitalWrite(in2, LOW);
    }
  }

  else
  {
    analogWrite(pwm, 0);

  }
}

// Scaling_____________________________________________________________________
float scalingFromPotToMm(int point)
{
  float scaled_target = float(point) * 140 / 1000;
  return scaled_target;
}

int scalingFromMmToPot(float point)
{
  int scaled_pot = int(point * 1000 / 140);
  return scaled_pot;
}

// Decode positions__________________________________________
float* decodePositions(String positions) {
  float positionList[6];
  String token = String(strtok(packet_buffer, ","));
  positionList[0] = token.toFloat();
  int i = 1;
  while (token != NULL) {
    token = strtok(NULL, ",");
    positionList[i] = token.toFloat();
    i += 1;
  }
  return positionList;
}

// Encode postions__________________________________________
String encodePositions(float* positions) {
  String positionString = "";
  for (int i = 0; i < 6; i++) {
    positionString += String(positions[i]);
    if (i != 5) {
      positionString += ",";
    }
  }
  return positionString;
}

// Convert the data from the sensor to newton based on SingleTact's calculations
float calculation(float sensorData)
{
  float calcData = ((sensorData - baselineOutput) / 512) * sensorRating;
  return calcData;
}

/* The function bellow is copied from SingleTact's Github:
   https://github.com/SingleTact/StandaloneArduino/blob/master/SingleTactDemo.ino.
*/
short readDataFromSensor(short address)
{
  byte i2cPacketLength = 6;//i2c packet length. Just need 6 bytes from each slave
  byte outgoingI2CBuffer[3];//outgoing array buffer
  byte incomingI2CBuffer[6];//incoming array buffer

  outgoingI2CBuffer[0] = 0x01;//I2c read command
  outgoingI2CBuffer[1] = 128;//Slave data offset
  outgoingI2CBuffer[2] = i2cPacketLength;//require 6 bytes

  Wire.beginTransmission(address); // transmit to device
  Wire.write(outgoingI2CBuffer, 3);// send out command
  byte error = Wire.endTransmission(); // stop transmitting and check slave status
  if (error != 0) return -1; //if slave not exists or has error, return -1
  Wire.requestFrom(address, i2cPacketLength);//require 6 bytes from slave

  byte incomeCount = 0;
  while (incomeCount < i2cPacketLength)    // slave may send less than requested
  {
    if (Wire.available())
    {
      incomingI2CBuffer[incomeCount] = Wire.read(); // receive a byte as character
      incomeCount++;
    }
    else
    {
      delayMicroseconds(10); //Wait 10us
    }
  }

  short rawData = (incomingI2CBuffer[4] << 8) + incomingI2CBuffer[5]; //get the raw data

  return rawData;
}

// LOOP_______________________________________________________________
void loop()
{
  int packet_size = udp_server.parsePacket();
  // If a packet is recieved
  if (packet_size)
  {
    for (int i = 0; i < UDP_TX_PACKET_MAX_SIZE; i++)
      packet_buffer[i] = 0;
    udp_server.read(packet_buffer, UDP_TX_PACKET_MAX_SIZE);

    if (String(packet_buffer) == "start") {
      startTalus = true;
    }
    else if (String(packet_buffer) == "stop") {
      startTalus = false;
    }
    // Takes the recieved leg lengths in millimeters, scales them and
    //puts them in a list
    else {
      float decoded_target[6];
      String servo = String(strtok(packet_buffer, ","));
      decoded_target[0] = servo.toFloat();
      int i = 1;

      while (servo != NULL) {
        servo = strtok(NULL, ",");
        decoded_target[i] = servo.toFloat();
        i += 1;
      }

      for (int i = 0; i < 6; i++) {
        servoPos[i] = scalingFromMmToPot(decoded_target[i]);
      }
    }
  }
  //Get actuator position
  int potPos[6];
  for (int i = 0; i < 6; i++)
  {
    potPos[i] = readPotentiometer(pot[i]);
  }
  // If it is the first loop, set the set-point for the actuators to
  //its current position
  if (oneshot) {
    for (int i = 0; i < 6; i++) {
      servoPos[i] = potPos[i];
    }

  }
  oneshot = false;

  // Every second, send the application the current leg lengths
  unsigned long runtime = millis() / 1000;
  if (runtime > interval) {
    float positions[6];
    for (int i = 0; i < 6; i++)
    {
      positions[i] = scalingFromPotToMm(potPos[i]);
    }
    String encoded_pos = encodePositions(positions);
    sendUDP(encoded_pos);
    interval += 1;
  }

  short data1 = readDataFromSensor(i2cAddress1); // talus sensor
  short data2 = readDataFromSensor(i2cAddress2); // side metatarsal sensor
  // Convert the data received from the sensor to newton
  float talus_data = calculation(data1);
  float metatarsal_data =  calculation(data2);

  // If the start button on the application has been pressed,
  //start force regulation and force logging
  if (startTalus == true) {
    float power_talus = calculateU(force_talus, talus_data, kp_talus_force,
    ki_talus_force, kd_talus_force, eprevTalus, eintegralTalus);
    setMotor(power_talus, PWM7, IN17, IN27);
    power_force = calculateU(force_metatarsal, metatarsal_data, kp_metatarsal,
    ki_metatarsal, kd_metatarsal, eprevMetatarsal, eintegralMetatarsal);
    if (millis() / 100 > interval2) {
      Serial.print(metatarsal_data);
      Serial.print(",");
      Serial.print(talus_data);
      Serial.print(",");
      Serial.println(millis());
      interval2 += 1;
    }
  }
  // If start has not been pressed or stop has been pressed,
  //retract the talus actuator and turn off the force regulation
  else {
    float power_talus = calculateU(30, analogRead(POT7), kp_talus_pos,
    ki_talus_pos, kd_talus_pos, eprevTalus, eintegralTalus);
    setMotor(power_talus, PWM7, IN17, IN27);
    power_force = 0;
  }

  // Calculate the gain and control each actuator
  for (int i = 0; i < 6; i++)
  {
    float power_platform = calculateU(servoPos[i], potPos[i], kp_platform,
    ki_platform, kd_platform, eprevList[i], eintegralList[i]);
    setMotor(power_platform - power_force, PWM[i], IN1[i], IN2[i]);

  }
}
