import numpy as np
def forwardkinematics(leglengths):
    #Constants
    baseAngles = [np.deg2rad(350), np.deg2rad(10), np.deg2rad(110), np.deg2rad(130),
                  np.deg2rad(230), np.deg2rad(250)]
    baseRadius = 84
    platformAngles = [np.deg2rad(310), np.deg2rad(50), np.deg2rad(70), np.deg2rad(170),
                      np.deg2rad(190), np.deg2rad(290)]
    platformRadius = 84
    basePos = np.zeros((6,3))
    platformPos = np.zeros((6,3))
    tolerance = 0.0000001
    iteration = 0
    maxIterations = 100000

    #calculating joint positions
    for i in range(6):
        basePos[i] = [baseRadius * np.cos(baseAngles[i]),
                      baseRadius * np.sin(baseAngles[i]), 0]
        platformPos[i] = [platformRadius * np.cos(platformAngles[i]),
                          platformRadius * np.sin(platformAngles[i]), 0]

    #initial guess
    initialguess = [0, 0, 300, np.deg2rad(0), np.deg2rad(0), np.deg2rad(0)]

    while iteration < maxIterations:

        yaw = initialguess[5]
        pitch = initialguess[4]
        roll = initialguess[3]
        rotationmatrix = np.array([[np.cos(yaw)*np.cos(pitch),
                                    np.cos(yaw)*np.sin(pitch)*np.sin(roll)-
                                    np.sin(yaw)*np.cos(roll),
                                    np.cos(yaw)*np.sin(pitch)*np.cos(roll)+
                                    np.sin(yaw)*np.sin(roll)],
                                   [np.sin(yaw)*np.cos(pitch),
                                    np.sin(yaw)*np.sin(pitch)*np.sin(roll)+
                                    np.cos(yaw)*np.cos(roll),
                                    np.sin(yaw)*np.sin(pitch)*np.cos(roll)-
                                    np.cos(yaw)*np.sin(roll)],
                                   [-np.sin(pitch),
                                    np.cos(pitch)*np.sin(roll),
                                    np.cos(pitch)*np.cos(roll)]])

        #Calculate leg length
        calcleglengthcords = np.zeros((6,3))
        calcleglength = np.zeros(6)
        for i in range(6):
            calcleglengthcords[i] = rotationmatrix@platformPos[i] +\
                                    initialguess[0:3] - basePos[i]
            calcleglength[i] = np.sqrt(calcleglengthcords[i][0]**2 +
                                       calcleglengthcords[i][1]**2 +
                                       calcleglengthcords[i][2]**2)

        #cost function
        cost = np.zeros((6,1))
        for i in range(6):
            cost[i][0] = calcleglength[i]**2 - leglengths[i]**2

        #return if cost function is below tolerance
        if abs(np.sum(cost)) < tolerance:
            initialguess[0] = round(initialguess[0], 2)
            initialguess[1] = round(initialguess[1], 2)
            initialguess[2] = round(initialguess[2], 2)
            initialguess[3] = round(np.rad2deg(initialguess[3]), 2)
            initialguess[4] = round(np.rad2deg(initialguess[4]), 2)
            initialguess[5] = round(np.rad2deg(initialguess[5]), 2)
            return initialguess

        #calculating partial derivatives
        dcostdx = np.zeros(6)
        dcostdy = np.zeros(6)
        dcostdz = np.zeros(6)
        dcostdyaw = np.zeros(6)
        dcostdpitch = np.zeros(6)
        dcostdroll = np.zeros(6)
        partialmatrix = np.zeros((6,6))
        for i in range(6):
            dcostdx[i] = 2*(initialguess[0] +
                            platformPos[i][0]*rotationmatrix[0][0] +
                            platformPos[i][1]*rotationmatrix[0][1] +
                            platformPos[i][2]*rotationmatrix[0][2]
                            - basePos[i][0])

            dcostdy[i] = 2*(initialguess[1] +
                            platformPos[i][0]*rotationmatrix[1][0] +
                            platformPos[i][1]*rotationmatrix[1][1] +
                            platformPos[i][2]*rotationmatrix[1][2]
                            - basePos[i][1])

            dcostdz[i] = 2*(initialguess[2] +
                            platformPos[i][0]*rotationmatrix[2][0] +
                            platformPos[i][1]*rotationmatrix[2][1] +
                            platformPos[i][2]*rotationmatrix[2][2]
                            - basePos[i][2])

            dcostdroll[i] = 2*(initialguess[0] - basePos[i][0])*\
                            (platformPos[i][1]*rotationmatrix[0][2] +
                            platformPos[i][2]*rotationmatrix[0][1]) +\
                            2*(initialguess[1] - basePos[i][1])*\
                        (platformPos[i][1]*rotationmatrix[1][2] +
                         platformPos[i][2]*rotationmatrix[1][1]) +\
                        2*(initialguess[2] - basePos[i][2])*\
                            (platformPos[i][1]*rotationmatrix[2][2] +
                            platformPos[i][2]*rotationmatrix[2][1])

            dcostdpitch[i] = 2*(initialguess[0] - basePos[i][0])*\
                             (-platformPos[i][0]*
                              np.sin(pitch)*np.cos(yaw) +
                                platformPos[i][1]*np.sin(roll)*
                              np.cos(pitch)*np.cos(yaw) +
                              platformPos[i][2]*np.cos(roll)*
                                np.cos(pitch)*np.cos(yaw)) +\
                             2*(initialguess[1] - basePos[i][1])*\
                             (-platformPos[i][0]*
                                np.sin(pitch)*np.sin(yaw) +
                              platformPos[i][1]*np.sin(roll)*
                              np.cos(pitch)*np.sin(yaw) +
                                platformPos[i][2]*np.cos(roll)*
                              np.cos(pitch)*np.sin(yaw)) -\
                             2*(initialguess[2] -
                                basePos[i][2])*(platformPos[i][0]*
                                                np.cos(pitch) +
                                                platformPos[i][1]*
                                                np.sin(roll)*
                                np.sin(pitch) +
                                                platformPos[i][2]*
                                                np.cos(roll)*
                                                np.sin(pitch))

            dcostdyaw[i] = -2*(initialguess[0] -
                               basePos[i][0])*(platformPos[i][0]*
                                               rotationmatrix[1][0] +
                            platformPos[i][1]*rotationmatrix[1][1] +
                                               platformPos[i][2]*
                                               rotationmatrix[1][2]) +\
                           2*(initialguess[1] - basePos[i][1])*\
                           (platformPos[i][0]*rotationmatrix[0][0] +
                            platformPos[i][1]*
                            rotationmatrix[0][1] +
                            platformPos[i][2]*rotationmatrix[0][2])

            partialmatrix[i][0] = dcostdx[i]
            partialmatrix[i][1] = dcostdy[i]
            partialmatrix[i][2] = dcostdz[i]
            partialmatrix[i][3] = dcostdroll[i]
            partialmatrix[i][4] = dcostdpitch[i]
            partialmatrix[i][5] = dcostdyaw[i]

        #update guess
        delta = np.linalg.inv(partialmatrix)@cost
        for i in range(6):
            initialguess[i] = initialguess[i] - delta[i][0]
        iteration +=1
    initialguess[0] = round(initialguess[0],2)
    initialguess[1] = round(initialguess[1],2)
    initialguess[2] = round(initialguess[2],2)
    initialguess[3] = round(np.rad2deg(initialguess[3]),2)
    initialguess[4] = round(np.rad2deg(initialguess[4]),2)
    initialguess[5] = round(np.rad2deg(initialguess[5]),2)
    return initialguess
