import numpy as np
def inversekinematics(desiredPosition):
    # Constants
    baseAngles = [np.deg2rad(350), np.deg2rad(10), np.deg2rad(110), np.deg2rad(130),
                  np.deg2rad(230), np.deg2rad(250)]
    baseRadius = 84
    platformAngles = [np.deg2rad(310), np.deg2rad(50), np.deg2rad(70),
                      np.deg2rad(170), np.deg2rad(190),
                      np.deg2rad(290)]
    platformRadius = 84
    basePos = np.zeros((6, 3))
    platformPos = np.zeros((6, 3))

    #Calculating joint positions
    for i in range(6):
        basePos[i] = [baseRadius * np.cos(baseAngles[i]),
                      baseRadius * np.sin(baseAngles[i]), 0]
        platformPos[i] = [platformRadius * np.cos(platformAngles[i]),
                          platformRadius * np.sin(platformAngles[i]), 0]

    #Rotation matrix
    yaw = np.deg2rad(desiredPosition[5])
    pitch = np.deg2rad(desiredPosition[4])
    roll = np.deg2rad(desiredPosition[3])
    rotationmatrix = np.array(
        [[np.cos(yaw) * np.cos(pitch),
          np.cos(yaw) * np.sin(pitch) * np.sin(roll) - np.sin(yaw) * np.cos(roll),
          np.cos(yaw) * np.sin(pitch) * np.cos(roll) + np.sin(yaw) * np.sin(roll)],
         [np.sin(yaw) * np.cos(pitch),
          np.sin(yaw) * np.sin(pitch) * np.sin(roll) + np.cos(yaw) * np.cos(roll),
          np.sin(yaw) * np.sin(pitch) * np.cos(roll) - np.cos(yaw) * np.sin(roll)],
         [-np.sin(pitch), np.cos(pitch) * np.sin(roll),
          np.cos(pitch) * np.cos(roll)]])

    #Calculating the leg length
    leglengthcords = np.zeros((6,3))
    leglength = np.zeros(6)
    for i in range(6):
        leglengthcords[i] = rotationmatrix @ platformPos[i] + desiredPosition[0:3]\
                            - basePos[i]
        leglength[i] = np.sqrt(leglengthcords[i][0] ** 2 + leglengthcords[i][1] ** 2 +
                               leglengthcords[i][2] ** 2)
    return leglength