from django.shortcuts import render, redirect
from django.http import JsonResponse
from django.core import serializers
from. import models
from. import kinematics
from. import sendPositions
from .forms import TreatmentForm
from .models import TreatmentPlan
from socket import *
import numpy as np
from. import inverseKinematics

#Function that is run when the user goes to the main page, the function starts
#communication with the Arduino and handles the treatment plan when it is uploaded
def indexAjax(request):
    kinematics.runkinematics()
    if request.method == 'POST':
        form = TreatmentForm(request.POST, request.FILES)
        if form.is_valid():
            form.save()
            latest = TreatmentPlan.objects.order_by('uploaded_at').last()
            sendPositions.sendpositions(str(latest.treatmentplan))
            return redirect('home')
    else:
        form = TreatmentForm()
    return render(request, "index.htm", {'form': form})

#Function used to continually update the values shown on the status page
def updateData(request):
    lengths = models.Leg.objects.all().values()
    position = models.platformPosition.objects.all().values()
    return JsonResponse({"lengths": list(lengths), "platformpos": list(position)},
                        status = 200)

#Function that is run when the status page is accessed, handles the
# manual control of the orthosis
def status(request):
    udp_socket = socket(AF_INET, SOCK_DGRAM)
    udp_socket.settimeout(1)
    arduino_ip = '192.168.10.240'
    arduino_port = 8888
    offset = -265
    lengths = models.Leg.objects.all().values()
    position = models.platformPosition.objects.all().values()
    if request.method == 'POST':
        if request.POST.get('legz', '') == '' and request.POST.get('xpos', '') == '':
            leg1 = float(request.POST.get('leg1pluss', 0))\
                   + float(request.POST.get('leg1minus', 0))\
                   + float(lengths[0]['length']) + offset
            leg2 = float(request.POST.get('leg2pluss', 0))\
                   + float(request.POST.get('leg2minus', 0))\
                   + float(lengths[1]['length']) + offset
            leg3 = float(request.POST.get('leg3pluss', 0))\
                   + float(request.POST.get('leg3minus', 0))\
                   + float(lengths[2]['length']) + offset
            leg4 = float(request.POST.get('leg4pluss', 0))\
                   + float(request.POST.get('leg4minus', 0))\
                   + float(lengths[3]['length']) + offset
            leg5 = float(request.POST.get('leg5pluss', 0))\
                   + float(request.POST.get('leg5minus', 0))\
                   + float(lengths[4]['length']) + offset
            leg6 = float(request.POST.get('leg6pluss', 0))\
                   + float(request.POST.get('leg6minus', 0))\
                   + float(lengths[5]['length']) + offset
            msg_to_encode = str(leg1)
            msg_to_encode += ","
            msg_to_encode += str(leg2)
            msg_to_encode += ","
            msg_to_encode += str(leg3)
            msg_to_encode += ","
            msg_to_encode += str(leg4)
            msg_to_encode += ","
            msg_to_encode += str(leg5)
            msg_to_encode += ","
            msg_to_encode += str(leg6)
            print(msg_to_encode)
            udp_socket.sendto(str(msg_to_encode).encode(), (arduino_ip, arduino_port))

        elif request.POST.get('xpos', '') == '':
            leg1 = float(request.POST.get('legz', '')) + float(lengths[0]['length'])\
                   + offset
            leg2 = float(request.POST.get('legz', '')) + float(lengths[1]['length'])\
                   + offset
            leg3 = float(request.POST.get('legz', '')) + float(lengths[2]['length'])\
                   + offset
            leg4 = float(request.POST.get('legz', '')) + float(lengths[3]['length'])\
                   + offset
            leg5 = float(request.POST.get('legz', '')) + float(lengths[4]['length'])\
                   + offset
            leg6 = float(request.POST.get('legz', '')) + float(lengths[5]['length'])\
                   + offset
            msg_to_encode = str(leg1)
            msg_to_encode += ","
            msg_to_encode += str(leg2)
            msg_to_encode += ","
            msg_to_encode += str(leg3)
            msg_to_encode += ","
            msg_to_encode += str(leg4)
            msg_to_encode += ","
            msg_to_encode += str(leg5)
            msg_to_encode += ","
            msg_to_encode += str(leg6)
            print(msg_to_encode)
            udp_socket.sendto(str(msg_to_encode).encode(), (arduino_ip, arduino_port))
        else:
            desiredpos = np.zeros(6)
            desiredpos[0] = float(request.POST.get('xpos', ''))
            desiredpos[1] = float(request.POST.get('ypos', ''))
            desiredpos[2] = float(request.POST.get('zpos', ''))
            desiredpos[3] = float(request.POST.get('roll', ''))
            desiredpos[4] = float(request.POST.get('pitch', ''))
            desiredpos[5] = float(request.POST.get('yaw', ''))
            calcpos = inverseKinematics.inversekinematics(desiredpos)
            leg1 = float(calcpos[0]) + offset
            leg2 = float(calcpos[1]) + offset
            leg3 = float(calcpos[2]) + offset
            leg4 = float(calcpos[3]) + offset
            leg5 = float(calcpos[4]) + offset
            leg6 = float(calcpos[5]) + offset
            msg_to_encode = str(leg1)
            msg_to_encode += ","
            msg_to_encode += str(leg2)
            msg_to_encode += ","
            msg_to_encode += str(leg3)
            msg_to_encode += ","
            msg_to_encode += str(leg4)
            msg_to_encode += ","
            msg_to_encode += str(leg5)
            msg_to_encode += ","
            msg_to_encode += str(leg6)
            udp_socket.sendto(str(msg_to_encode).encode(), (arduino_ip, arduino_port))
    return render(request, "status.html", {"lengths": list(lengths),
                                           "platformpos": list(position)})

#Function that is run when either the start or stop button is pressed,
#sends the command to the Arduino
def encode(request):
    udp_socket = socket(AF_INET, SOCK_DGRAM)
    udp_socket.settimeout(1)
    arduino_ip = '192.168.10.240'
    arduino_port = 8888
    if request.method == 'POST':
        if request.POST.get('start', '') == "start":
            msg_to_encode = request.POST.get('start', '')
            udp_socket.sendto(str(msg_to_encode).encode(), (arduino_ip, arduino_port))
            sendPositions.startButton = True
        if request.POST.get('stop', '') == "stop":
            msg_to_encode = request.POST.get('stop', '')
            udp_socket.sendto(str(msg_to_encode).encode(), (arduino_ip, arduino_port))
            sendPositions.startButton = False
    return render(request, 'index.htm', {})